@@ -20,6 +20,7 @@
 
 #include <Nux/Nux.h>
 #include <NuxCore/Logger.h>
+#include <UnityCore/GLibWrapper.h>
 #include <boost/algorithm/string/erase.hpp>
 
 #include "PanelMenuView.h"
@@ -27,6 +28,7 @@
 #include "unity-shared/CairoTexture.h"
 #include "unity-shared/DecorationStyle.h"
 #include "unity-shared/PanelStyle.h"
+#include "unity-shared/RawPixel.h"
 #include "unity-shared/UnitySettings.h"
 #include "unity-shared/UBusMessages.h"
 #include "unity-shared/UScreen.h"
@@ -43,15 +45,14 @@ DECLARE_LOGGER(logger, "unity.panel.menu");
 
 namespace
 {
-  const int MAIN_LEFT_PADDING = 4;
-  const int TITLE_PADDING = 2;
-  const int MENUBAR_PADDING = 4;
+  const RawPixel MAIN_LEFT_PADDING = 4_em;
+  const RawPixel TITLE_PADDING = 2_em;
+  const RawPixel MENUBAR_PADDING = 4_em;
   const int MENU_ENTRIES_PADDING = 6;
 
   const std::string NEW_APP_HIDE_TIMEOUT = "new-app-hide-timeout";
   const std::string NEW_APP_SHOW_TIMEOUT = "new-app-show-timeout";
   const std::string WINDOW_MOVED_TIMEOUT = "window-moved-timeout";
-  const std::string WINDOW_ACTIVATED_TIMEOUT = "window-activated-timeout";
   const std::string UPDATE_SHOW_NOW_TIMEOUT = "update-show-now-timeout";
   const std::string INTEGRATED_MENUS_DOUBLE_CLICK_TIMEOUT = "integrated-menus-double-click-timeout";
 
@@ -83,14 +84,15 @@ std::string get_current_desktop()
 }
 
 PanelMenuView::PanelMenuView(menu::Manager::Ptr const& menus)
-  : menu_manager_(menus)
-  , matcher_(bamf_matcher_get_default())
+  : active_window(0)
+  , maximized_window(0)
+  , focused(true)
+  , menu_manager_(menus)
   , is_inside_(false)
   , is_grabbed_(false)
   , is_maximized_(false)
   , is_desktop_focused_(false)
   , last_active_view_(nullptr)
-  , new_application_(nullptr)
   , switcher_showing_(false)
   , spread_showing_(false)
   , launcher_keynav_(false)
@@ -100,23 +102,21 @@ PanelMenuView::PanelMenuView(menu::Manager::Ptr const& menus)
   , ignore_menu_visibility_(false)
   , integrated_menus_(menu_manager_->integrated_menus())
   , always_show_menus_(menu_manager_->always_show_menus())
-  , active_xid_(0)
+  , ignore_leave_events_(false)
   , desktop_name_(get_current_desktop())
 {
-  BamfWindow* active_win = bamf_matcher_get_active_window(matcher_);
-  if (BAMF_IS_WINDOW(active_win))
-    active_xid_ = bamf_window_get_xid(active_win);
+  if (ApplicationWindowPtr const& win = ApplicationManager::Default().GetActiveWindow())
+    active_window = win->window_id();
 
-  SetupPanelMenuViewSignals();
   SetupWindowButtons();
   SetupTitlebarGrabArea();
+  SetupPanelMenuViewSignals();
   SetupWindowManagerSignals();
   SetupUBusManagerInterests();
 
   opacity = 0.0f;
 
-  if (Refresh())
-    FullRedraw();
+  RefreshAndRedraw();
 }
 
 PanelMenuView::~PanelMenuView()
@@ -128,8 +128,10 @@ PanelMenuView::~PanelMenuView()
 void PanelMenuView::OnStyleChanged()
 {
   int height = panel::Style::Instance().PanelHeight(monitor_);
+  double scale = Settings::Instance().em(monitor_)->DPIScale();
   window_buttons_->SetMinimumHeight(height);
   window_buttons_->SetMaximumHeight(height);
+  window_buttons_->SetLeftAndRightPadding(MAIN_LEFT_PADDING.CP(scale), MENUBAR_PADDING.CP(scale));
   window_buttons_->UpdateDPIChanged();
 
   layout_->SetLeftAndRightPadding(window_buttons_->GetContentWidth(), 0);
@@ -140,14 +142,13 @@ void PanelMenuView::OnStyleChanged()
 
 void PanelMenuView::SetupPanelMenuViewSignals()
 {
-  active_win_changed_signal_.Connect(matcher_, "active-window-changed",
-                                     sigc::mem_fun(this, &PanelMenuView::OnActiveWindowChanged));
-  active_app_changed_signal_.Connect(matcher_, "active-application-changed",
-                                     sigc::mem_fun(this, &PanelMenuView::OnActiveAppChanged));
-  view_opened_signal_.Connect(matcher_, "view-opened",
-                              sigc::mem_fun(this, &PanelMenuView::OnViewOpened));
-  view_closed_signal_.Connect(matcher_, "view-closed",
-                              sigc::mem_fun(this, &PanelMenuView::OnViewClosed));
+  auto& am = ApplicationManager::Default();
+  am.active_window_changed.connect(sigc::mem_fun(this, &PanelMenuView::OnActiveWindowChanged));
+  am.active_application_changed.connect(sigc::mem_fun(this, &PanelMenuView::OnActiveAppChanged));
+  am.application_started.connect(sigc::mem_fun(this, &PanelMenuView::OnApplicationStarted));
+  am.application_stopped.connect(sigc::mem_fun(this, &PanelMenuView::OnApplicationClosed));
+  am.window_opened.connect(sigc::mem_fun(this, &PanelMenuView::OnWindowOpened));
+  am.window_closed.connect(sigc::mem_fun(this, &PanelMenuView::OnWindowClosed));
 
   mouse_enter.connect(sigc::mem_fun(this, &PanelMenuView::OnPanelViewMouseEnter));
   mouse_leave.connect(sigc::mem_fun(this, &PanelMenuView::OnPanelViewMouseLeave));
@@ -157,6 +158,20 @@ void PanelMenuView::SetupPanelMenuViewSignals()
 
   menu_manager_->integrated_menus.changed.connect(sigc::mem_fun(this, &PanelMenuView::OnLIMChanged));
   menu_manager_->always_show_menus.changed.connect(sigc::mem_fun(this, &PanelMenuView::OnAlwaysShowMenusChanged));
+
+  auto update_target_cb = sigc::hide(sigc::mem_fun(this, &PanelMenuView::UpdateTargetWindowItems));
+  maximized_window.changed.connect(update_target_cb);
+  active_window.changed.connect(update_target_cb);
+
+  focused.changed.connect([this] (bool focused) {
+    Refresh(true);
+    window_buttons_->focused = focused;
+
+    for (auto const& entry : entries_)
+      entry.second->SetFocusedState(focused);
+
+    FullRedraw();
+  });
 }
 
 void PanelMenuView::SetupWindowButtons()
@@ -164,7 +179,6 @@ void PanelMenuView::SetupWindowButtons()
   window_buttons_ = new WindowButtons();
   window_buttons_->SetParentObject(this);
   window_buttons_->monitor = monitor_;
-  window_buttons_->controlled_window = active_xid_;
   window_buttons_->opacity = 0.0f;
   window_buttons_->SetLeftAndRightPadding(MAIN_LEFT_PADDING, MENUBAR_PADDING);
   window_buttons_->SetMaximumHeight(panel::Style::Instance().PanelHeight(monitor_));
@@ -208,15 +222,18 @@ void PanelMenuView::SetupWindowManagerSignals()
   wm.window_unminimized.connect(sigc::mem_fun(this, &PanelMenuView::OnWindowUnminimized));
   wm.window_maximized.connect(sigc::mem_fun(this, &PanelMenuView::OnWindowMaximized));
   wm.window_restored.connect(sigc::mem_fun(this, &PanelMenuView::OnWindowRestored));
+  wm.window_fullscreen.connect(sigc::mem_fun(this, &PanelMenuView::OnWindowMaximized));
+  wm.window_unfullscreen.connect(sigc::mem_fun(this, &PanelMenuView::OnWindowUnFullscreen));
   wm.window_unmapped.connect(sigc::mem_fun(this, &PanelMenuView::OnWindowUnmapped));
   wm.window_mapped.connect(sigc::mem_fun(this, &PanelMenuView::OnWindowMapped));
   wm.window_moved.connect(sigc::mem_fun(this, &PanelMenuView::OnWindowMoved));
   wm.window_resized.connect(sigc::mem_fun(this, &PanelMenuView::OnWindowMoved));
   wm.initiate_spread.connect(sigc::mem_fun(this, &PanelMenuView::OnSpreadInitiate));
   wm.terminate_spread.connect(sigc::mem_fun(this, &PanelMenuView::OnSpreadTerminate));
-  wm.initiate_expo.connect(sigc::mem_fun(this, &PanelMenuView::OnExpoInitiate));
-  wm.terminate_expo.connect(sigc::mem_fun(this, &PanelMenuView::OnExpoTerminate));
-  wm.screen_viewport_switch_ended.connect(sigc::mem_fun(this, &PanelMenuView::OnExpoTerminate));
+  wm.initiate_expo.connect(sigc::mem_fun(this, &PanelMenuView::RefreshAndRedraw));
+  wm.terminate_expo.connect(sigc::mem_fun(this, &PanelMenuView::RefreshAndRedraw));
+  wm.screen_viewport_switch_ended.connect(sigc::mem_fun(this, &PanelMenuView::RefreshAndRedraw));
+  wm.show_desktop_changed.connect(sigc::mem_fun(this, &PanelMenuView::OnShowDesktopChanged));
 }
 
 void PanelMenuView::SetupUBusManagerInterests()
@@ -252,6 +269,37 @@ void PanelMenuView::AddIndicator(indicator::Indicator::Ptr const& indicator)
   PanelIndicatorsView::AddIndicator(indicator);
 }
 
+void PanelMenuView::UpdateTargetWindowItems()
+{
+  Window old_target = window_buttons_->controlled_window;
+  Window target_window = integrated_menus_ ? maximized_window() : active_window();
+
+  if (old_target != target_window)
+  {
+    window_buttons_->controlled_window = target_window;
+
+    if (ApplicationWindowPtr const& win = ApplicationManager::Default().GetWindowForId(target_window))
+    {
+      auto refresh_cb = sigc::hide(sigc::mem_fun(this, &PanelMenuView::RefreshAndRedraw));
+      win_name_changed_conn_ = win->title.changed.connect(refresh_cb);
+
+      if (ApplicationPtr const& app = win->application())
+        app_name_changed_conn_ = app->title.changed.connect(refresh_cb);
+    }
+
+    ClearEntries();
+
+    if (indicator::AppmenuIndicator::Ptr appmenu = menu_manager_->AppMenu())
+    {
+      for (auto const& entry : appmenu->GetEntriesForWindow(target_window))
+        OnEntryAdded(entry);
+    }
+  }
+
+  if (integrated_menus_)
+    focused = (target_window == active_window());
+}
+
 void PanelMenuView::FullRedraw()
 {
   QueueDraw();
@@ -261,14 +309,16 @@ void PanelMenuView::FullRedraw()
 void PanelMenuView::OnLIMChanged(bool lim)
 {
   integrated_menus_ = lim;
-  new_application_ = nullptr;
+  new_application_.reset();
 
   if (!integrated_menus_)
   {
     CheckMouseInside();
-    window_buttons_->focused = true;
+    focused = true;
   }
 
+  UpdateTargetWindowItems();
+
   Refresh(true);
   FullRedraw();
 }
@@ -280,7 +330,7 @@ void PanelMenuView::OnAlwaysShowMenusChanged(bool always_show_menus)
   if (!always_show_menus_)
     CheckMouseInside();
 
-  FullRedraw();
+  QueueDraw();
 }
 
 nux::Area* PanelMenuView::FindAreaUnderMouse(const nux::Point& mouse_position, nux::NuxEventType event_type)
@@ -302,7 +352,7 @@ nux::Area* PanelMenuView::FindAreaUnderMouse(const nux::Point& mouse_position, n
       return titlebar_grab_area_.GetPointer();
   }
 
-  if (is_maximized_ || spread_showing_ || (integrated_menus_ && GetMaximizedWindow() != 0))
+  if (is_maximized_ || spread_showing_ || (integrated_menus_ && maximized_window() != 0))
   {
     found_area = window_buttons_->FindAreaUnderMouse(mouse_position, event_type);
     NUX_RETURN_VALUE_IF_NOTNULL(found_area, found_area);
@@ -313,7 +363,7 @@ nux::Area* PanelMenuView::FindAreaUnderMouse(const nux::Point& mouse_position, n
     found_area = titlebar_grab_area_->FindAreaUnderMouse(mouse_position, event_type);
     NUX_RETURN_VALUE_IF_NOTNULL(found_area, found_area);
 
-    if (integrated_menus_ && GetMaximizedWindow() != 0)
+    if (integrated_menus_ && maximized_window() != 0)
     {
       /* When the integrated menus are enabled, that area must act both like an
        * indicator-entry view and like a panel-grab-area, so not to re-implement
@@ -382,10 +432,7 @@ void PanelMenuView::OnFadeAnimatorUpdated(double progress)
 
 bool PanelMenuView::ShouldDrawMenus() const
 {
-  if (integrated_menus_ && !is_maximized_)
-    return false;
-
-  if (we_control_active_ && !switcher_showing_ && !launcher_keynav_ && !ignore_menu_visibility_ && !entries_.empty())
+  if ((we_control_active_ || integrated_menus_) && !switcher_showing_ && !launcher_keynav_ && !ignore_menu_visibility_ && HasVisibleMenus())
   {
     WindowManager& wm = WindowManager::Default();
 
@@ -410,7 +457,7 @@ bool PanelMenuView::ShouldDrawButtons() const
   if (integrated_menus_)
   {
     if (!WindowManager::Default().IsExpoActive())
-      return (GetMaximizedWindow() != 0);
+      return (maximized_window() != 0);
 
     return false;
   }
@@ -677,7 +724,7 @@ void PanelMenuView::DrawContent(nux::GraphicsEngine& GfxContext, bool force_draw
 
     if (new_application_ && !is_inside_)
     {
-      if (opacity() != 1.0f)
+      if (opacity() != 1.0f && menu_manager_->discovery() > 0)
         StartFadeIn(menu_manager_->discovery_fadein());
     }
     else
@@ -724,41 +771,13 @@ void PanelMenuView::DrawContent(nux::GraphicsEngine& GfxContext, bool force_draw
 std::string PanelMenuView::GetActiveViewName(bool use_appname) const
 {
   std::string label;
-  BamfWindow* window;
+  auto& am = ApplicationManager::Default();
 
-  window = bamf_matcher_get_active_window(matcher_);
-
-  if (BAMF_IS_WINDOW(window))
+  if (ApplicationWindowPtr const& window = am.GetActiveWindow())
   {
-    BamfView *view = reinterpret_cast<BamfView*>(window);
-    Window window_xid = bamf_window_get_xid(window);
+    Window window_xid = window->window_id();
 
-    if (bamf_window_get_window_type(window) == BAMF_WINDOW_DOCK)
-    {
-      auto panel = const_cast<PanelMenuView*>(this)->GetTopLevelViewWindow();
-      if (static_cast<nux::BaseWindow*>(panel)->GetInputWindowId() == window_xid)
-        return desktop_name_;
-
-      std::vector<Window> const& our_xids = nux::XInputWindow::NativeHandleList();
-
-      if (std::find(our_xids.begin(), our_xids.end(), window_xid) != our_xids.end())
-      {
-        /* If the active window is an unity window, we need to fallback to the
-         * top one, anyway we should always avoid to focus unity internal windows */
-        BamfWindow* top_win = GetBamfWindowForXid(GetTopWindow());
-
-        if (top_win && top_win != window)
-        {
-          window = top_win;
-        }
-        else
-        {
-          return "";
-        }
-      }
-    }
-
-    if (bamf_window_get_window_type(window) == BAMF_WINDOW_DESKTOP)
+    if (window->type() == WindowType::DESKTOP)
     {
       label = desktop_name_;
     }
@@ -767,28 +786,17 @@ std::string PanelMenuView::GetActiveViewName(bool use_appname) const
       return "";
     }
 
-    if (WindowManager::Default().IsWindowMaximized(window_xid) && !use_appname)
-    {
-      label = glib::String(bamf_view_get_name(view)).Str();
-    }
+    if (window_xid == maximized_window() && !use_appname)
+      label = window->title();
 
     if (label.empty())
     {
-      BamfApplication* app;
-      app = bamf_matcher_get_application_for_window(matcher_, window);
-
-      if (BAMF_IS_APPLICATION(app))
-      {
-        view = reinterpret_cast<BamfView*>(app);
-        label = glib::String(bamf_view_get_name(view)).Str();
-      }
+      if (ApplicationPtr const& app = am.GetActiveApplication())
+        label = app->title();
     }
 
     if (label.empty())
-    {
-      view = reinterpret_cast<BamfView*>(window);
-      label = glib::String(bamf_view_get_name(view)).Str();
-    }
+      label = window->title();
   }
 
   return label;
@@ -796,25 +804,17 @@ std::string PanelMenuView::GetActiveViewName(bool use_appname) const
 
 std::string PanelMenuView::GetMaximizedViewName(bool use_appname) const
 {
-  Window maximized = GetMaximizedWindow();
-  BamfWindow* window = nullptr;
+  Window maximized = maximized_window();
   std::string label;
 
-  window = GetBamfWindowForXid(maximized);
-
-  if (BAMF_IS_WINDOW(window))
+  if (ApplicationWindowPtr const& window = ApplicationManager::Default().GetWindowForId(maximized))
   {
-    BamfView* view = reinterpret_cast<BamfView*>(window);
-    label = glib::String(bamf_view_get_name(view)).Str();
+    label = window->title();
 
     if (use_appname || label.empty())
     {
-      BamfApplication* app = bamf_matcher_get_application_for_window(matcher_, window);
-
-      if (BAMF_IS_APPLICATION(app))
-        view = reinterpret_cast<BamfView*>(app);
-
-      label = glib::String(bamf_view_get_name(view)).Str();
+      if (ApplicationPtr const& app = window->application())
+        label = app->title();
     }
   }
 
@@ -830,18 +830,18 @@ void PanelMenuView::UpdateTitleTexture(nux::Geometry const& geo, std::string con
   auto const& style = decoration::Style::Get();
   auto text_size = style->TitleNaturalSize(label);
   auto state = WidgetState::NORMAL;
-  float dpi_scale = Settings::Instance().em(monitor_)->DPIScale();
+  double dpi_scale = Settings::Instance().em(monitor_)->DPIScale();
 
   if (integrated_menus_ && !is_desktop_focused_ && !WindowManager::Default().IsExpoActive())
   {
     title_geo_.x = geo.x + window_buttons_->GetBaseWidth() + (style->TitleIndent() * dpi_scale);
 
-    if (!window_buttons_->focused())
+    if (!focused())
       state = WidgetState::BACKDROP;
   }
   else
   {
-    title_geo_.x = geo.x + (MAIN_LEFT_PADDING + TITLE_PADDING) * dpi_scale;
+    title_geo_.x = geo.x + MAIN_LEFT_PADDING.CP(dpi_scale) + TITLE_PADDING.CP(dpi_scale);
   }
 
   title_geo_.y = geo.y + (geo.height - (text_size.height * dpi_scale)) / 2;
@@ -904,20 +904,9 @@ bool PanelMenuView::Refresh(bool force)
   if (geo.width > monitor_geo_.width)
     return false;
 
-  if (integrated_menus_)
-  {
-    Window maximized = GetMaximizedWindow();
-    window_buttons_->controlled_window = maximized;
-    window_buttons_->focused = (active_xid_ == maximized);
-  }
-  else
-  {
-    window_buttons_->controlled_window = active_xid_;
-  }
-
   std::string const& new_title = GetCurrentTitle();
 
-  if (new_title == panel_title_ && !force && last_geo_ == geo && title_texture_)
+  if (!force && new_title == panel_title_ && last_geo_ == geo && title_texture_)
   {
     // No need to redraw the title, let's save some CPU time!
     return false;
@@ -936,6 +925,12 @@ bool PanelMenuView::Refresh(bool force)
   return true;
 }
 
+void PanelMenuView::RefreshAndRedraw()
+{
+  if (Refresh())
+    QueueDraw();
+}
+
 void PanelMenuView::OnActiveChanged(PanelIndicatorEntryView* view, bool is_active)
 {
   if (is_active)
@@ -950,24 +945,25 @@ void PanelMenuView::OnActiveChanged(PanelIndicatorEntryView* view, bool is_activ
     }
   }
 
-  if (Refresh())
-    FullRedraw();
+  RefreshAndRedraw();
 }
 
 void PanelMenuView::OnEntryAdded(indicator::Entry::Ptr const& entry)
 {
-  PanelIndicatorEntryView* view;
+  auto parent_window = entry->parent_window();
+  Window target = integrated_menus_ ? maximized_window() : active_window();
 
-  view = new PanelIndicatorEntryView(entry, MENU_ENTRIES_PADDING, IndicatorEntryType::MENU);
-  entry->show_now_changed.connect(sigc::mem_fun(this, &PanelMenuView::UpdateShowNow));
-  AddEntryView(view);
+  if (!parent_window || parent_window == target)
+    AddEntryView(new PanelIndicatorEntryView(entry, MENU_ENTRIES_PADDING, IndicatorEntryType::MENU));
 }
 
 void PanelMenuView::OnEntryViewAdded(PanelIndicatorEntryView* view)
 {
+  view->SetFocusedState(focused());
   view->mouse_enter.connect(sigc::mem_fun(this, &PanelMenuView::OnPanelViewMouseEnter));
   view->mouse_leave.connect(sigc::mem_fun(this, &PanelMenuView::OnPanelViewMouseLeave));
   view->active_changed.connect(sigc::mem_fun(this, &PanelMenuView::OnActiveChanged));
+  view->show_now_changed.connect(sigc::mem_fun(this, &PanelMenuView::UpdateShowNow));
 }
 
 void PanelMenuView::NotifyAllMenusClosed()
@@ -983,22 +979,13 @@ void PanelMenuView::NotifyAllMenusClosed()
   }
 }
 
-void PanelMenuView::OnNameChanged(BamfView* bamf_view, gchar* new_name, gchar* old_name)
-{
-  if (Refresh())
-    FullRedraw();
-}
-
 bool PanelMenuView::OnNewAppShow()
 {
-  BamfApplication* active_app = bamf_matcher_get_active_application(matcher_);
-  new_application_ = glib::Object<BamfApplication>(active_app, glib::AddRef());
+  new_application_ = ApplicationManager::Default().GetActiveApplication();
   QueueDraw();
 
   if (sources_.GetSource(NEW_APP_HIDE_TIMEOUT))
-  {
     new_app_menu_shown_ = false;
-  }
 
   auto cb_func = sigc::mem_fun(this, &PanelMenuView::OnNewAppHide);
   sources_.AddTimeoutSeconds(menu_manager_->discovery(), cb_func, NEW_APP_HIDE_TIMEOUT);
@@ -1015,70 +1002,61 @@ bool PanelMenuView::OnNewAppHide()
   return false;
 }
 
-void PanelMenuView::OnViewOpened(BamfMatcher *matcher, BamfView *view)
+void PanelMenuView::OnApplicationStarted(ApplicationPtr const& app)
 {
   /* FIXME: here we should also check for if the view is also user_visible
    * but it seems that BAMF doesn't handle this correctly after some
    * stress tests (repeated launches). */
-  if (!BAMF_IS_APPLICATION(view) || integrated_menus_)
+  if (integrated_menus_)
     return;
 
-  new_apps_.push_front(glib::Object<BamfApplication>(BAMF_APPLICATION(view), glib::AddRef()));
+  new_apps_.push_front(app);
 }
 
-void PanelMenuView::OnApplicationClosed(BamfApplication* app)
+void PanelMenuView::OnApplicationClosed(ApplicationPtr const& app)
 {
-  if (BAMF_IS_APPLICATION(app) && !integrated_menus_)
+  if (app && !integrated_menus_)
   {
     if (std::find(new_apps_.begin(), new_apps_.end(), app) != new_apps_.end())
     {
-      new_apps_.remove(glib::Object<BamfApplication>(app, glib::AddRef()));
+      new_apps_.remove(app);
     }
     else if (new_apps_.empty())
     {
-      new_application_ = nullptr;
+      new_application_.reset();
     }
   }
 
   if (app == new_application_)
   {
-    new_application_ = nullptr;
+    new_application_.reset();
   }
 }
 
-void PanelMenuView::OnViewClosed(BamfMatcher *matcher, BamfView *view)
+void PanelMenuView::OnWindowOpened(ApplicationWindowPtr const& win)
 {
-  if (reinterpret_cast<BamfView*>(view_name_changed_signal_.object()) == view)
+  if (win->window_id() == window_buttons_->controlled_window() &&
+      win->title.changed.empty())
   {
-    view_name_changed_signal_.Disconnect();
-  }
-
-  if (BAMF_IS_APPLICATION(view))
-  {
-    OnApplicationClosed(reinterpret_cast<BamfApplication*>(view));
-  }
-  else if (reinterpret_cast<BamfApplication*>(view) == new_application_)
-  {
-    new_application_ = nullptr;
-  }
-  else if (BAMF_IS_WINDOW(view))
-  {
-    /* FIXME, this can be removed when window_unmapped WindowManager signal
-     * will emit the proper xid */
-    Window xid = bamf_window_get_xid(reinterpret_cast<BamfWindow*>(view));
-    OnWindowUnmapped(xid);
+    /* This is a not so nice workaround that we need to include here, since
+     * BAMF might be late in informing us about a new window, and thus we
+     * can't connect to it's signals (as not available in the App Manager). */
+    window_buttons_->controlled_window = 0;
+    UpdateTargetWindowItems();
   }
 }
 
-void PanelMenuView::OnActiveAppChanged(BamfMatcher *matcher,
-                                       BamfApplication* old_app,
-                                       BamfApplication* new_app)
+void PanelMenuView::OnWindowClosed(ApplicationWindowPtr const& win)
 {
-  if (BAMF_IS_APPLICATION(new_app))
-  {
-    app_name_changed_signal_.Connect(BAMF_VIEW(new_app), "name-changed",
-                                     sigc::mem_fun(this, &PanelMenuView::OnNameChanged));
+  /* FIXME, this can be removed when window_unmapped WindowManager signal
+   * will emit the proper xid */
+  OnWindowUnmapped(win->window_id());
+}
 
+void PanelMenuView::OnActiveAppChanged(ApplicationPtr const& new_app)
+{
+  if (new_app)
+  {
     if (integrated_menus_ || always_show_menus_)
       return;
 
@@ -1111,64 +1089,40 @@ void PanelMenuView::OnActiveAppChanged(BamfMatcher *matcher,
   }
 }
 
-void PanelMenuView::OnActiveWindowChanged(BamfMatcher *matcher, BamfView* old_view, BamfView* new_view)
+void PanelMenuView::OnActiveWindowChanged(ApplicationWindowPtr const& new_win)
 {
   show_now_activated_ = false;
   is_maximized_ = false;
   is_desktop_focused_ = false;
-  active_xid_ = 0;
-  bool force_refresh = false;
+  Window active_xid = 0;
 
   sources_.Remove(WINDOW_MOVED_TIMEOUT);
 
-  if (BAMF_IS_WINDOW(new_view))
+  if (new_win)
   {
-    BamfWindow* window = reinterpret_cast<BamfWindow*>(new_view);
-    active_xid_ = bamf_window_get_xid(window);
-    is_maximized_ = (bamf_window_maximized(window) == BAMF_WINDOW_MAXIMIZED);
+    active_xid = new_win->window_id();
+    is_maximized_ = new_win->maximized() || WindowManager::Default().IsWindowFullscreen(active_xid);
 
-    if (bamf_window_get_window_type(window) == BAMF_WINDOW_DESKTOP)
+    if (new_win->type() == WindowType::DESKTOP)
     {
-      is_desktop_focused_ = !GetMaximizedWindow();
+      is_desktop_focused_ = !maximized_window();
       we_control_active_ = true;
     }
     else
     {
-      we_control_active_ = IsWindowUnderOurControl(active_xid_);
+      we_control_active_ = IsWindowUnderOurControl(active_xid);
     }
 
     if (is_maximized_)
     {
-      maximized_wins_.erase(std::remove(maximized_wins_.begin(), maximized_wins_.end(), active_xid_), maximized_wins_.end());
-      maximized_wins_.push_front(active_xid_);
+      maximized_wins_.erase(std::remove(maximized_wins_.begin(), maximized_wins_.end(), active_xid), maximized_wins_.end());
+      maximized_wins_.push_front(active_xid);
+      UpdateMaximizedWindow();
     }
-
-    // register callback for new view
-    view_name_changed_signal_.Connect(new_view, "name-changed",
-                                      sigc::mem_fun(this, &PanelMenuView::OnNameChanged));
-
-    if (integrated_menus_)
-      force_refresh = is_maximized_;
   }
 
-  if (!force_refresh && BAMF_IS_WINDOW(old_view) && integrated_menus_)
-    force_refresh = (bamf_window_maximized(reinterpret_cast<BamfWindow*>(old_view)) == BAMF_WINDOW_MAXIMIZED);
-
-  if (ShouldDrawMenus())
-  {
-    // Wait 100ms before showing the menus again if we've just switched view
-    // this is done because the menus are updated by the indicator with some
-    // delay, and we don't want to see the previous menus and then the new ones
-    ignore_menu_visibility_ = true;
-    sources_.AddTimeout(100, [this] {
-      ignore_menu_visibility_ = false;
-      QueueDraw();
-      return false;
-    }, WINDOW_ACTIVATED_TIMEOUT);
-  }
-
-  if (Refresh(force_refresh))
-    FullRedraw();
+  active_window = active_xid;
+  RefreshAndRedraw();
 }
 
 void PanelMenuView::OnSpreadInitiate()
@@ -1183,53 +1137,44 @@ void PanelMenuView::OnSpreadTerminate()
   QueueDraw();
 }
 
-void PanelMenuView::OnExpoInitiate()
-{
-  if (Refresh())
-    QueueDraw();
-}
-
-void PanelMenuView::OnExpoTerminate()
-{
-  if (Refresh())
-    QueueDraw();
-}
-
 void PanelMenuView::OnWindowMinimized(Window xid)
 {
   maximized_wins_.erase(std::remove(maximized_wins_.begin(), maximized_wins_.end(), xid), maximized_wins_.end());
+  UpdateMaximizedWindow();
 
-  if (xid == active_xid_)
+  if (xid == active_window())
   {
-    if (Refresh())
-      QueueDraw();
+    RefreshAndRedraw();
   }
   else if (integrated_menus_ && window_buttons_->controlled_window == xid)
   {
-    if (Refresh())
-      QueueDraw();
+    RefreshAndRedraw();
   }
 }
 
 void PanelMenuView::OnWindowUnminimized(Window xid)
 {
-  if (xid == active_xid_)
+  if (xid == active_window())
   {
     if (WindowManager::Default().IsWindowMaximized(xid))
+    {
       maximized_wins_.push_front(xid);
+      UpdateMaximizedWindow();
+    }
 
-    if (Refresh())
-      QueueDraw();
+    RefreshAndRedraw();
   }
   else
   {
     if (WindowManager::Default().IsWindowMaximized(xid))
+    {
       maximized_wins_.push_back(xid);
+      UpdateMaximizedWindow();
+    }
 
     if (integrated_menus_ && IsWindowUnderOurControl(xid))
     {
-      if (Refresh())
-        QueueDraw();
+      RefreshAndRedraw();
     }
   }
 }
@@ -1239,16 +1184,15 @@ void PanelMenuView::OnWindowUnmapped(Window xid)
   // FIXME: compiz doesn't give us a valid xid (is always 0 on unmap)
   // we need to do this again on BamfView closed signal.
   maximized_wins_.erase(std::remove(maximized_wins_.begin(), maximized_wins_.end(), xid), maximized_wins_.end());
+  UpdateMaximizedWindow();
 
-  if (xid == active_xid_)
+  if (xid == active_window())
   {
-    if (Refresh())
-      QueueDraw();
+    RefreshAndRedraw();
   }
   else if (integrated_menus_ && window_buttons_->controlled_window == xid)
   {
-    if (Refresh())
-      QueueDraw();
+    RefreshAndRedraw();
   }
 }
 
@@ -1256,77 +1200,87 @@ void PanelMenuView::OnWindowMapped(Window xid)
 {
   if (WindowManager::Default().IsWindowMaximized(xid))
   {
-    if (xid == active_xid_)
+    if (xid == active_window())
     {
       maximized_wins_.push_front(xid);
+      UpdateMaximizedWindow();
 
-      if (Refresh())
-        QueueDraw();
+      RefreshAndRedraw();
     }
     else
     {
       maximized_wins_.push_back(xid);
+      UpdateMaximizedWindow();
     }
   }
 }
 
 void PanelMenuView::OnWindowMaximized(Window xid)
 {
-  if (xid == active_xid_)
+  if (xid == active_window())
   {
     maximized_wins_.push_front(xid);
+    UpdateMaximizedWindow();
 
     // We need to update the is_inside_ state in the case of maximization by grab
     CheckMouseInside();
     is_maximized_ = true;
 
-    if (Refresh())
-      FullRedraw();
+    RefreshAndRedraw();
   }
   else
   {
     maximized_wins_.push_back(xid);
+    UpdateMaximizedWindow();
 
     if (integrated_menus_ && IsWindowUnderOurControl(xid))
     {
-      if (Refresh())
-        QueueDraw();
+      RefreshAndRedraw();
     }
   }
 }
 
+void PanelMenuView::OnWindowUnFullscreen(Window xid)
+{
+  if (!WindowManager::Default().IsWindowMaximized(xid))
+    OnWindowRestored(xid);
+}
+
 void PanelMenuView::OnWindowRestored(Window xid)
 {
   maximized_wins_.erase(std::remove(maximized_wins_.begin(), maximized_wins_.end(), xid), maximized_wins_.end());
+  UpdateMaximizedWindow();
 
-  if (active_xid_ == xid)
+  if (active_window() == xid)
   {
     is_maximized_ = false;
     is_grabbed_ = false;
-
-    if (Refresh())
-      FullRedraw();
+    RefreshAndRedraw();
   }
   else if (integrated_menus_ && window_buttons_->controlled_window == xid)
   {
-    if (Refresh())
-      QueueDraw();
+    RefreshAndRedraw();
   }
 }
 
+void PanelMenuView::OnShowDesktopChanged()
+{
+  UpdateMaximizedWindow();
+}
+
+
 bool PanelMenuView::UpdateActiveWindowPosition()
 {
-  bool we_control_window = IsWindowUnderOurControl(active_xid_);
+  bool we_control_window = IsWindowUnderOurControl(active_window);
 
   if (we_control_window != we_control_active_)
   {
     we_control_active_ = we_control_window;
 
-    if (!entries_.empty())
+    if (HasVisibleMenus())
       on_indicator_updated.emit();
 
-    if (Refresh())
-      QueueDraw();
+    RefreshAndRedraw();
   }
 
   return false;
@@ -1334,21 +1288,17 @@ bool PanelMenuView::UpdateActiveWindowPosition()
 
 void PanelMenuView::OnWindowMoved(Window xid)
 {
-  if (active_xid_ == xid)
+  if (!integrated_menus_ && active_window() == xid && UScreen::GetDefault()->GetMonitors().size() > 1)
   {
     /* When moving the active window, if the current panel is controlling
      * the active window, then we postpone the timeout function every movement
      * that we have, setting a longer timeout.
-     * Otherwise, if the moved window is not controlled by the current panel
+     * Otherwise, if the movedPanelMenuView::OnWindowMovedPanelMenuView::OnWindowMoved window is not controlled by the current panel
      * every few millisecond we check the new window position */
 
     unsigned int timeout_length = 250;
 
-    if (we_control_active_)
-    {
-      sources_.Remove(WINDOW_MOVED_TIMEOUT);
-    }
-    else
+    if (!we_control_active_)
     {
       if (sources_.GetSource(WINDOW_MOVED_TIMEOUT))
         return;
@@ -1359,6 +1309,9 @@ void PanelMenuView::OnWindowMoved(Window xid)
     auto cb_func = sigc::mem_fun(this, &PanelMenuView::UpdateActiveWindowPosition);
     sources_.AddTimeout(timeout_length, cb_func, WINDOW_MOVED_TIMEOUT);
   }
+
+  if (std::find(maximized_wins_.begin(), maximized_wins_.end(), xid) != maximized_wins_.end())
+    UpdateMaximizedWindow();
 }
 
 bool PanelMenuView::IsWindowUnderOurControl(Window xid) const
@@ -1391,7 +1344,7 @@ bool PanelMenuView::IsValidWindow(Window xid) const
   return false;
 }
 
-Window PanelMenuView::GetMaximizedWindow() const
+void PanelMenuView::UpdateMaximizedWindow()
 {
   Window window_xid = 0;
 
@@ -1406,71 +1359,39 @@ Window PanelMenuView::GetMaximizedWindow() const
     }
   }
 
-  return window_xid;
+  maximized_window = window_xid;
 }
 
 Window PanelMenuView::GetTopWindow() const
 {
   Window window_xid = 0;
-  GList* windows = bamf_matcher_get_window_stack_for_monitor(matcher_, monitor_);
 
-  for (GList* l = windows; l; l = l->next)
+  for (auto const& win : ApplicationManager::Default().GetWindowsForMonitor(monitor_))
   {
-    if (!BAMF_IS_WINDOW(l->data))
-      continue;
+    Window xid = win->window_id();
 
-    Window xid = bamf_window_get_xid(static_cast<BamfWindow*>(l->data));
-    bool visible = bamf_view_is_user_visible(static_cast<BamfView*>(l->data));
-
-    if (visible && IsValidWindow(xid))
+    if (win->visible() && IsValidWindow(xid))
     {
       window_xid = xid;
     }
   }
 
-  g_list_free(windows);
-
   return window_xid;
 }
 
-BamfWindow* PanelMenuView::GetBamfWindowForXid(Window xid) const
-{
-  BamfWindow* window = nullptr;
-
-  if (xid != 0)
-  {
-    GList* windows = bamf_matcher_get_windows(matcher_);
-
-    for (GList* l = windows; l; l = l->next)
-    {
-      if (!BAMF_IS_WINDOW(l->data))
-        continue;
-
-      auto win = static_cast<BamfWindow*>(l->data);
-
-      if (bamf_window_get_xid(win) == xid)
-      {
-        window = win;
-        break;
-      }
-    }
-
-    g_list_free(windows);
-  }
-
-  return window;
-}
-
 void PanelMenuView::ActivateIntegratedMenus(nux::Point const& click)
 {
   if (!layout_->GetAbsoluteGeometry().IsInside(click))
     return;
 
-  unsigned double_click_wait = Settings::Instance().lim_double_click_wait();
+  auto& settings = Settings::Instance();
 
-  if (double_click_wait > 0)
+  if (!focused && !settings.lim_unfocused_popup())
+    return;
+
+  if (settings.lim_double_click_wait() > 0)
   {
-    sources_.AddTimeout(double_click_wait, [this, click] {
+    sources_.AddTimeout(settings.lim_double_click_wait(), [this, click] {
       ActivateEntryAt(click.x, click.y);
       return false;
     }, INTEGRATED_MENUS_DOUBLE_CLICK_TIMEOUT);
@@ -1489,15 +1410,18 @@ void PanelMenuView::ActivateIntegratedMenus(nux::Point const& click)
 
 void PanelMenuView::OnMaximizedActivate(int x, int y)
 {
-  Window maximized = GetMaximizedWindow();
+  Window maximized = maximized_window();
 
   if (maximized != 0)
   {
-    if (maximized != active_xid_)
+    if (maximized != active_window())
     {
-      WindowManager::Default().Activate(maximized);
+      auto& wm = WindowManager::Default();
+      wm.Raise(maximized);
+      wm.Activate(maximized);
     }
-    else if (integrated_menus_)
+
+    if (integrated_menus_)
     {
       // Adjusting the click coordinates to be absolute.
       auto const& grab_geo = titlebar_grab_area_->GetAbsoluteGeometry();
@@ -1509,7 +1433,7 @@ void PanelMenuView::OnMaximizedActivate(int x, int y)
 
 void PanelMenuView::MaximizedWindowWMAction(int x, int y, unsigned button)
 {
-  Window maximized = GetMaximizedWindow();
+  Window maximized = maximized_window();
 
   if (!maximized)
     return;
@@ -1590,7 +1514,7 @@ void PanelMenuView::OnMaximizedGrabStart(int x, int y)
    * This is a workaround to avoid that the grid plugin would be fired
    * showing the window shape preview effect. See bug #838923 */
 
-  Window maximized = GetMaximizedWindow();
+  Window maximized = maximized_window();
 
   if (maximized != 0)
   {
@@ -1611,7 +1535,7 @@ void PanelMenuView::OnMaximizedGrabMove(int x, int y)
   x += titlebar_grab_area_->GetAbsoluteX();
   y += titlebar_grab_area_->GetAbsoluteY();
 
-  Window maximized = GetMaximizedWindow();
+  Window maximized = maximized_window();
 
   /* When the drag goes out from the Panel, start the real movement.
    *
@@ -1649,8 +1573,7 @@ void PanelMenuView::OnMaximizedGrabMove(int x, int y)
 
       is_inside_ = true;
       is_grabbed_ = true;
-      if (Refresh())
-        FullRedraw();
+      RefreshAndRedraw();
 
       /* Ungrab the pointer and start the X move, to make the decorator handle it */
       titlebar_grab_area_->SetGrabbed(false);
@@ -1670,8 +1593,7 @@ void PanelMenuView::OnMaximizedGrabEnd(int x, int y)
   if (!is_inside_)
     is_grabbed_ = false;
 
-  if (Refresh())
-    FullRedraw();
+  RefreshAndRedraw();
 }
 
 // Introspectable
@@ -1686,6 +1608,8 @@ void PanelMenuView::AddProperties(debug::IntrospectionData& introspection)
   PanelIndicatorsView::AddProperties(introspection);
 
   introspection
+  .add("focused", focused())
+  .add("integrated_menus", integrated_menus_)
   .add("mouse_inside", is_inside_)
   .add("always_show_menus", always_show_menus_)
   .add("grabbed", is_grabbed_)
@@ -1694,7 +1618,8 @@ void PanelMenuView::AddProperties(debug::IntrospectionData& introspection)
   .add("panel_title", panel_title_)
   .add("desktop_active", (panel_title_ == desktop_name_))
   .add("monitor", monitor_)
-  .add("active_window", active_xid_)
+  .add("active_window", active_window())
+  .add("maximized_window", maximized_window())
   .add("draw_menus", ShouldDrawMenus())
   .add("draw_window_buttons", ShouldDrawButtons())
   .add("controls_active_window", we_control_active_)
@@ -1730,8 +1655,7 @@ void PanelMenuView::OnSwitcherShown(GVariant* data)
     show_now_activated_ = false;
   }
 
-  if (Refresh())
-    QueueDraw();
+  RefreshAndRedraw();
 }
 
 void PanelMenuView::OnLauncherKeyNavStarted(GVariant* data)
@@ -1752,9 +1676,7 @@ void PanelMenuView::OnLauncherKeyNavEnded(GVariant* data)
 
   launcher_keynav_ = false;
   CheckMouseInside();
-
-  if (Refresh())
-    QueueDraw();
+  RefreshAndRedraw();
 }
 
 void PanelMenuView::OnLauncherSelectionChanged(GVariant* data)
@@ -1821,52 +1743,40 @@ void PanelMenuView::SetMonitor(int monitor)
 
   maximized_wins_.clear();
   monitor_geo_ = UScreen::GetDefault()->GetMonitorGeometry(monitor_);
-  GList* windows = bamf_matcher_get_window_stack_for_monitor(matcher_, monitor_);
 
-  for (GList* l = windows; l; l = l->next)
+  for (auto const& win : ApplicationManager::Default().GetWindowsForMonitor(monitor_))
   {
-    if (!BAMF_IS_WINDOW(l->data))
-      continue;
+    auto xid = win->window_id();
 
-    auto window = static_cast<BamfWindow*>(l->data);
-    auto view = static_cast<BamfView*>(l->data);
-    auto xid = bamf_window_get_xid(window);
+    if (win->active())
+      active_window = xid;
 
-    if (bamf_view_is_active(view))
-      active_xid_ = xid;
-
-    if (bamf_window_maximized(window) == BAMF_WINDOW_MAXIMIZED)
+    if (win->maximized() || WindowManager::Default().IsWindowFullscreen(xid))
     {
-      if (xid == active_xid_)
+      if (win->active())
         maximized_wins_.push_front(xid);
       else
         maximized_wins_.push_back(xid);
     }
   }
 
-  Window maximized = GetMaximizedWindow();
-  Window buttons_win = 0;
-
-  if (integrated_menus_)
-  {
-    buttons_win = maximized;
-    window_buttons_->focused = (maximized == active_xid_);
-  }
-  else
-  {
-    buttons_win = (maximized == active_xid_) ? maximized : 0;
-  }
-
   window_buttons_->monitor = monitor_;
-  window_buttons_->controlled_window = buttons_win;
+  UpdateMaximizedWindow();
 
   OnStyleChanged();
-  g_list_free(windows);
 }
 
 bool PanelMenuView::HasMenus() const
 {
-  if (entries_.empty())
+  if (!HasVisibleMenus())
+    return false;
+
+  return integrated_menus_ || we_control_active_;
+}
+
+bool PanelMenuView::HasKeyActivableMenus() const
+{
+  if (!HasVisibleMenus())
     return false;
 
   return integrated_menus_ ? is_maximized_ : we_control_active_;
@@ -1904,9 +1814,14 @@ void PanelMenuView::OnPanelViewMouseEnter(int x, int y, unsigned long mouse_butt
   }
 }
 
+void PanelMenuView::IgnoreLeaveEvents(bool ignore)
+{
+  ignore_leave_events_ = ignore;
+}
+
 void PanelMenuView::OnPanelViewMouseLeave(int x, int y, unsigned long mouse_button_state, unsigned long special_keys_state)
 {
-  if (always_show_menus_)
+  if (always_show_menus_ || ignore_leave_events_)
     return;
 
   if (is_inside_)
